/* sig-gen.cpp by Adam Somers for Music 256a hw1 */

#include <iostream>

#include "AudioServer.h"
#include "SignalGenerators.h"

#include "chuck_fft.h"
#include "marsyas/AutoCorrelation.h"

// OpenGL
#if defined(__MACOSX_CORE__)
#include <GLUT/glut.h>
#include <OpenGL/gl.h>
#include <OpenGL/glu.h>
#else
#include <GL/glut.h>
#include <GL/gl.h>
#include <GL/glu.h>
#endif

// gl callbacks
void idleFunc( );
void displayFunc( );
void reshapeFunc( int width, int height );
void keyboardFunc( unsigned char, int, int );
void mouseFunc( int button, int state, int x, int y );
void motionFunc(int x, int y);
void timerCallback(int v);
void initialize_graphics( );
void createMenu(void);
void menu(int value);

void DrawString(int x,int y, std::string& s, void* font, float scaleFactor = 0.2f);
void cycleColors();
void go2d();
void go3d();

enum ColorScheme
{
   kMidnight = 0,
   kNeon,
   
   kNumColorSchemes
};


const unsigned qMidnightColors[] = { 0x0EBF7F, 0x419979, 0x62E5B6, 0x47A684, 0x377F65 };
const unsigned qNeonColors[] =     {0xFF0000, 0x00FF00, 0x0000FF, 0xFF00FF, 0x00FFFF};

const unsigned* qColorSchemes[5]  = { qMidnightColors, qNeonColors };

unsigned gColorScheme = 0;

// width and height of the window
GLsizei gWidth = 800;
GLsizei gHeight = 600;
GLsizei gLastWidth = gWidth;
GLsizei gLastHeight = gHeight;

GLenum gFillmode = GL_FILL;
GLfloat gLight0Pos[4] = { 2.0f, 1.2f, 4.0f, 1.0f };
GLfloat gLight1Ambient[] = { .2f, .2f, .2f, 1.0f };
GLfloat gLight1Diffuse[] = { 1.0f, 1.0f, 1.0f, 1.0f };
GLfloat gLight1Specular[] = { 1.0f, 1.0f, 1.0f, 1.0f };
GLfloat gLight1Pos[4] = { -2.0f, 0.0f, -4.0f, 1.0f };
// modelview stuff
GLfloat gAngleY = 0.0f;
GLfloat gAngleX = 0.0f;
GLfloat gInc = 0.0f;
GLfloat gEyeY = 0.0f;
GLboolean gDisplay = TRUE;
GLboolean gFullscreen = FALSE;

const int gBufferSize = 1024;

float gFrameRate = 40; // frames/sec
float gRefreshRate = 1000 / gFrameRate; // ms
float gSamplesPerFrame = 12000;

float gBuff[gBufferSize];
float gWindowBuffer[gBufferSize];
float gFFTBuffer[gBufferSize];
int gBigBin = 0;
// stores min/max peak pairs
SampleAccumulator::PeakBuffer gPeaks;
// obtains peak data from the signal chain
SampleAccumulator gAccumulator;

int gMenuItem = 0;
GLfloat gInc_val_mouse = 360 / gFrameRate;

bool gLeftButtondown = false;
int gMouseOriginX = 0;
int gMouseOriginY = 0;

// experimental
int gHopCount = 0;
bool gRings = false;
int gHop = 10;

bool gTriggered = false;
bool gStaggered = false;
float gFFTPitch = 0;
float gLogFactor = 1;
bool gShowHelp = true;

AutoCorrelation gAutoCorrelation(gBufferSize);

enum Mode
{
   kScroll = 0,
   kScope,
   kSpectrum,
   
   kNumModes
};

class GLDisplay
{
public:
   GLDisplay(float x, float y, float w, float h)
   : fX(x)
   , fY(y)
   , fW(w)
   , fH(h)
   , fZ(0)
   , fR(1.0)
   , fG(0)
   , fB(0)
   , fA(1.1)
   , fVisible(true)
   , fInitialZAngle(0)
   {
   }
   
   void Display()
   {
      glPushMatrix();
      if (gStaggered)
         glTranslatef(fX, fY / 2.f, fZ / 2.f);
      else
         glTranslatef(fX, fY, fZ);
      glScalef(fW,fH, 1);
      glColor4f(fR, fG, fB, fA);
      Draw();
      glPopMatrix();
   }
   
   virtual void Draw() = 0;
   
   void SetY(float y)
   {
      fY=y;
   }
   
   void SetH(float h)
   {
      fH = h;
   }
   
   void SetZ(float z)
   {
      fZ = z;
   }
   
   void SetInitialZAngle(float phi)
   {
      fInitialZAngle = phi;
   }
   
   void SetColor(float r, float g, float b, float a)
   {
      fR = r; fG = g; fB = b; fA = a;
   }
   
   void SetVisible(bool visible)
   {
      fVisible = visible;
   }
   
   bool Visible() const { return fVisible; }
   
protected:
   float fX, fY, fW, fH, fZ, fR, fG, fB, fA;
   bool fVisible;
   float fInitialZAngle;
};

class Spheres : public GLDisplay
{
public:
   Spheres(int x, int y, int w, int h)
   : GLDisplay(x,y,w,h)
   {
   }
   
   void Draw()
   {
      glPushMatrix();
      
      glRotatef( gAngleY / 2.f, 0.0f, 1.0f, 0.0);
      glLineWidth(1);
      glutSolidSphere(.5, 20, 20);
      
      glTranslatef( .75, 0, 0);
      glPushMatrix();
      glRotatef(45, 1, 0, 0.0);
      glRotatef( gAngleY, 0, 0, 1);
      glColor4f(1.0, 0.0, 1.0, 1.0);
      glutSolidSphere(.25, 20, 20);
      glPopMatrix();
      
      glRotatef(gAngleY, 1, 0,0);
      glTranslatef( 0, .375, 0);
      glColor4f(1.0, 1.0, 0.0, 1.0);
      glutSolidSphere(.125, 20, 20);
      glPopMatrix();
   }
};

class WaveScroll : public GLDisplay
{
public:
   WaveScroll(float x, float y, float w, float h)
   : GLDisplay(x,y,w,h)
   {
   }
   
   void Draw()
   {
      glPushMatrix();
      glRotatef(-gAngleX, 1, 0, 0);
      
      
      
      // get the current buffer's peak data
      SampleAccumulator::PeakBuffer peaks = gAccumulator.Get();
      while (!peaks.empty())
      {
         gPeaks.push_back(peaks.front());
         peaks.pop_front();
      }
      
      while (gPeaks.size() > gSamplesPerFrame)
         gPeaks.pop_front();
      
      float x = -1;
      float xInc = fabs((2*x) / (float)gPeaks.size());
   	      
      glLineWidth(1);
      
      //glRotatef( gAngleY, 0, 1.0f, 0.0);
      //glRotatef( gAngleX, 1, 0, 0.0);
      
      //glutWireCube(2);
      
      if (!gRings)
      {
         glBegin( GL_LINE_STRIP );
         SampleAccumulator::PeakBuffer::iterator i;
         for (i = gPeaks.begin(); i != gPeaks.end(); ++i)
         {
            
            
            float y0 = (*i).first;
            float y1 = (*i).second;
            
            glVertex3f(x, y0, 0);
            glVertex3f(x, y1, 0);
            
            x+=xInc;
         }
         glEnd();
      }
      else
      {
         SampleAccumulator::PeakBuffer::iterator i;
         for (i = gPeaks.begin(); i != gPeaks.end(); ++i)
         {
            float y0 = (*i).first;
            float y1 = (*i).second;
            glBegin( GL_LINE_STRIP );
            for(int i=0;i<35;i++){
               float xcoord = y1 * cos(i*2*PI/10.f);
               float ycoord = y1 * sin(i*2*PI/10.f);	
               glVertex3f(x, ycoord, xcoord);
            }
            x+=xInc;
            glEnd();
         }
      }
      
      
      glPopMatrix( );   
      
   }
};

class Scope : public GLDisplay
{
public:
   Scope(float x, float y, float w, float h)
   : GLDisplay(x,y,w,h)
   {
   }
   
   void Draw()
   {
      int delay = 0;
      
      for (int i = 0; i < gBufferSize - 1; ++i)
      {
         if (gBuff[i] < 0.0001 && gBuff[i+1] >= 0.0001)
         {
            gTriggered = true;
            break;
         }
         delay++;
      }
      
      if (!gTriggered)
         return;
      
      glPushMatrix();
      glRotatef(-gAngleX, 1, 0, 0);
      
      float x = -1;
      float xInc =fabs((2*x) / (float)gBufferSize);
            
      glLineWidth(1);
      
      //glRotatef( gAngleY, 0, 1.0f, 0.0);
      //glRotatef( gAngleX, 1, 0, 0.0);
      
      //glutWireCube(2);
      
      if (!gRings)
      {
         glBegin( GL_LINE_STRIP );
         for (int i = 0; i < gBufferSize; ++i)
         {
            if (i < delay)
               continue;
            
            float y = gBuff[i];
            glVertex3f(x, y, 0);
            x+=xInc;
         }
         glEnd();
      }
      else
      {
         
         for (int i = 0; i < gBufferSize; ++i)
         {
            if (i < delay)
               continue;
            
            float y = gBuff[i];
            
            if (!gRings)
            {
               glVertex3f(x, y, 0);
            }
            else
            {
               glBegin( GL_LINE_LOOP );	
               for(int i=0;i<35;i++){
                  float xcoord = y * cos(i*2*PI/10.f);
                  float ycoord = y * sin(i*2*PI/10.f);	
                  glVertex3f(x, ycoord, xcoord);
               }        
               glEnd();
            }
            x+=xInc;
         }
      }
      
      glPopMatrix( );
      
      gTriggered = false;
   }
};

class Spectrum : public GLDisplay
{
public:
   Spectrum(float x, float y, float w, float h)
   : GLDisplay(x,y,w,h)
   {
      fAnalysisBuffer = new float[gBufferSize];
      memset(fAnalysisBuffer, 0.f, gBufferSize * sizeof(float));
      fWindowBuffer = new float[gBufferSize];
      hanning(fWindowBuffer, gBufferSize);
   }
   
   void Draw()
   {
      glPushMatrix();
      glRotatef(-gAngleX, 1, 0, 0);
      
      float x = -1;
      float xInc =fabs((2*x) / (float)gBufferSize) * 2;
      
      int points = gBufferSize/2.f;
      float xPos[points];
      for (int i = 0; i < points; ++i)
      {
         xPos[i] = ::powf(i / (float)points, gLogFactor);
         xPos[i] -= .49;
         xPos[i] *= 2;
      }
      
      glLineWidth(1);
      
      //glRotatef( gAngleY, 0, 1.0f, 0.0);
      //glRotatef( gAngleX, 1, 0, 0.0);
      
      //glutWireCube(2);
      
      //glTranslatef(0, -fH/4.f,0); // shift sectrum down since it only displays positive values
      
      memcpy( fAnalysisBuffer, gBuff, gBufferSize * sizeof(float));
      
      apply_window( (float*)fAnalysisBuffer, fWindowBuffer, gBufferSize );
      rfft( (float *)fAnalysisBuffer, gBufferSize/2, FFT_FORWARD );
      // cast to complex
      complex * cbuf = (complex *)fAnalysisBuffer;
     
      gBigBin = 0;
      float max = 0;
      glBegin( GL_LINES );
      for (int i = 0; i < gBufferSize/2; ++i)
      {
         float y = ::pow( 25 * cmp_abs( cbuf[i] ), .5 );
         if (y > max) {
            max = y;
            gBigBin = i;
         }
         glVertex3f(xPos[i], 0, 0);
         glVertex3f(xPos[i], y, 0);
         x+=xInc;
      }
      glEnd();
      
      gFFTPitch = gBigBin * 44100 / (float)gBufferSize;
      
      glPopMatrix( );	
      
   }
   
   void Get(float* dest, int size)
   {
      complex * cbuf = (complex *)fAnalysisBuffer;
      for (int i = 0; i < size; ++i)
      {
         dest[i] = ::pow( 25 * cmp_abs( cbuf[i] ), .5 );
      }
   }
   
private:
   float* fAnalysisBuffer;
   float* fWindowBuffer;
};

class Waterfall : public GLDisplay
{
public:
   Waterfall(float x, float y, float w, float h)
   : GLDisplay(x,y,w,h)
   , fEchos(50)
   , fR(0)
   , fW(fEchos - 1)
   , fFade(true)
   {
      fSpectrum = new Spectrum(x,y,w,h);
      
      fEcho = new float*[fEchos];
      for (int i = 0; i < fEchos; ++i)
      {
         fEcho[i] = new float[gBufferSize / 2];
         memset(fEcho[i], 0.f, (gBufferSize / 2) * sizeof(float));
      }
   }
   
   ~Waterfall()
   {
      delete fSpectrum;
      delete[] fEcho;
   }
   
   void Draw()
   {
   
      //glTranslatef(0,-1, 0);
      //glScalef(1,2, 1);
      
      glRotatef(fInitialZAngle, 1, 0, 0);
      
      glColor4f(1,0,0,0);
      fSpectrum->Draw();      
      
      fSpectrum->Get(fEcho[fR], gBufferSize / 2);
      
      int points = gBufferSize/2.f;
      float xPos[points];
      for (int i = 0; i < points; ++i)
      {
         xPos[i] = ::powf(i / (float)points, gLogFactor);
         xPos[i] -= .5;
         xPos[i] *= 2;
      }
      
      for (unsigned i = 0; i < fEchos; ++i)
      {
         float x = -1;
         float xInc =fabs((2*x) / (float)gBufferSize) * 2;
         

         
         glTranslatef(0, 1.f / fEchos, 0);
         
         glPushMatrix();
         
         glColor4f(1,0,0,1- (fFade ? (i / (float)fEchos) : 0));
         
         glRotatef(-gAngleX - fInitialZAngle, 1, 0, 0);
         
         glBegin(GL_LINE_STRIP);
         
         for (int j = 0; j < gBufferSize / 2; ++j)
         {
            //printf("%d\n", (fR - i) % (fEchos - 1));
            int index = fR - i;
            if (index < 0)
               index += fEchos;
            glVertex2f(xPos[j], fEcho[index][j]);
            x+=xInc;
         }

         glEnd();
         
         glPopMatrix();
      }
      
      //fR++;
      fR = (fR + 1) % (fEchos);
      
   }
   
   void SetFade(bool fade)
   {
      fFade = fade;
   }
   
private:
   Spectrum* fSpectrum;
   float** fEcho;
   unsigned int fEchos;
   int fW;
   int fR;
   bool fFade;
};

class Widget
{
public:
   Widget(int x, int y, int w, int h)
   : fX(x)
   , fY(y)
   , fW(w)
   , fH(h)
   {
   }
   
   bool Hit(int x, int y)
   {
      
      if ((x >= fX && x <= fX + fW) && (gHeight - y >= fY && gHeight - y <= fY + fH))
      {  
         return true;
      }
      else
      {
         return false;
      }
      
   }
   
   int X() const { return fX; }
   int Y() const { return fY; }
   int W() const { return fW; }
   int H() const { return fH; }
   
protected:
   int fX, fY, fW, fH;
};

class Slider : public Widget
{
public:
   Slider(int x, int y, int w, int h)
   : Widget(x,y,w,h)
   {
      fTouched = false;
      fVal = 0.f;
   }
   
   void Draw()
   {
      go2d();
      
      glPushMatrix();
      glLineWidth(2);
      glColor4f(1,1,1,1);
      glBegin(GL_LINE_STRIP);
      glVertex2f(fX + fVal * fW,  fY + 2);
      glVertex2f(fX + fVal * fW, fY + fH - 2);
      glEnd();
      glPopMatrix();
      
      glColor4f(0,0,0,1);
      glRecti(fX+2, fY+2, fX + fW - 2, fY + fH - 2);
      glColor4f(1,0,0,1);
      glRecti(fX, fY, fX + fW, fY + fH);
      
      
      
      //glScalef(1, fW/fH,1);
      
      //glPushMatrix();
      //glTranslatef(gMouseOriginX/(float)gWidth, 0, 0);
      //glutWireSphere(fH, 5, 3);
      //glPopMatrix();
      
      go3d();
   }
   

   
   bool Touched() const { return fTouched; }
   
   void Touch(int x, int y)
   {
      fTouched = true;
      
      int localX = std::max(x - fX, 0);
      
      fVal = std::min(localX, fW) / (float)fW;
   }
   
   void UnTouch()
   {
      fTouched = false;
   }
   
   void SetValue(float val)
   {
      fVal = val;
      if (fVal < 0) val = 0;
      if (fVal > 1) val = 1;
   }
   
   float Value() const { return fVal; }
   
private:
   float fVal;
   bool fTouched;
};


class ToggleButton : public Widget
{
public:
   ToggleButton(int x, int y, int w, int h)
   : Widget(x,y,w,h)
   , fToggled(false)
   {
   }
   
   void Draw()
   {
      go2d();

      if (fToggled)
      {
         glPushMatrix();
         glColor4f(1,1,1,1);
         glBegin(GL_LINES);
         glVertex2f(fX+2,  fY+2);
         glVertex2f(fX + fW - 2,  fY + fH - 2);
         glVertex2f(fX+2,  fY + fH-2);
         glVertex2f(fX + fW - 2,  fY + 2);
         glEnd();
         glPopMatrix();
      }

      glColor4f(0,0,0,1);
      glRecti(fX+2, fY+2, fX + fW - 2, fY + fH - 2);
      glColor4f(1,0,0,1);
      glRecti(fX, fY, fX + fW, fY + fH);

      go3d();
   }
   
   void Toggle()
   {
      fToggled = !fToggled;
   }
   
   bool Toggled() const { return fToggled; }
   
private:
   bool fToggled;
};

Spheres gSpheres(0,-1,1,1);
WaveScroll gWaveScroll(0,-.5f,1,1);
Scope gScope(0,0,1,1);
Spectrum gSpectrum(0,.5f,1,1);
Waterfall gWaterfall(0,.5f,1,1);
Slider gLoggynessSlider(20,20,100,20);
Slider gScrollSizeSlider(170, 20, 100, 20);
ToggleButton gToggle(300, 20, 20, 20);

std::deque<GLDisplay*> gDisplays;

void UpdateControls()
{
   //gStaggered = gToggle.Toggled();
   gLogFactor = (1 - gLoggynessSlider.Value()) * 0.75 + 0.25;
   gSamplesPerFrame = gScrollSizeSlider.Value() * 40000 + 512;
}

void go2d() {
   
   glMatrixMode(GL_PROJECTION);
   glPushMatrix();
   glLoadIdentity();
   gluOrtho2D(0.0, gWidth, 0.0, gHeight);
   
   glMatrixMode(GL_MODELVIEW);
   glPushMatrix();
   glLoadIdentity();
   
   glDisable(GL_TEXTURE);
   glDisable(GL_TEXTURE_2D);   
   
}

void go3d() {
   glEnable(GL_TEXTURE);
   glEnable(GL_TEXTURE_2D);
   
   glMatrixMode(GL_MODELVIEW);
   glPopMatrix();
   
   glMatrixMode(GL_PROJECTION);
   glPopMatrix();
   
   reshapeFunc(gWidth, gHeight);
}


int main(int argc, char** argv)
{
   memset(gBuff, 0.f, gBufferSize * sizeof(float));
   memset(gFFTBuffer, 0.f, gBufferSize * sizeof(float));

   for (int i = 0; i < gSamplesPerFrame; ++i)
   {
      gPeaks.push_back(std::make_pair(0,0));
   }
	
   // make the transform window
   hanning( gWindowBuffer, gBufferSize );

	std::cout << "Hit q to quit.\n";
	
   RtAudioDriver driver(gBufferSize);    // init rtaudio
   
   InputSource i;          // connect input to output through muted accumulator
   gAccumulator.SetInput(&i);
   Multiplier m(&gAccumulator,NULL,0.f);
   AudioServer::GetInstance()->AddClient(&m, 0);
   
   gDisplays.push_back(&gScope);
   gDisplays.push_back(&gWaveScroll);
   gDisplays.push_back(&gWaterfall);
   gDisplays.push_back(&gSpectrum);
   
   //gDisplays.push_back(&gSpheres);
   
   gLoggynessSlider.SetValue(0.2);
   gScrollSizeSlider.SetValue(0.5);
   
   
   cycleColors();
   

#ifdef __OS_MACOSX__
   //save working dir
   char * cwd = getcwd( NULL, 0 );
#endif
   // initialize GLUT
   glutInit( &argc, argv );
  
#ifdef __OS_MACOSX__
   //restore working dir
   chdir( cwd );
   free( cwd );
#endif
  
   glutInitDisplayMode( GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH );
   glutInitWindowSize( gWidth, gHeight );
   glutInitWindowPosition( 100, 100 );
   glutCreateWindow( "SoundPrism" );
   //glutIdleFunc( idleFunc );
   glutDisplayFunc( displayFunc );
   glutReshapeFunc( reshapeFunc );
   glutKeyboardFunc( keyboardFunc );
   glutMouseFunc( mouseFunc );
   glutMotionFunc(motionFunc);        
   glutTimerFunc(gRefreshRate,timerCallback,0);
   createMenu();
   if( gFullscreen )
      glutFullScreen();
   
   // do our own initialization
   initialize_graphics( );

   glutMainLoop();
	
   for(;;)
      sleep(1);

   return 0;
   }


void initialize_graphics()
{
   // set the GL clear color - use when the color buffer is cleared
   glClearColor( 0.0f, 0.0f,0.0f, 1.0f );
   // set the shading model to 'smooth'
   glShadeModel( GL_SMOOTH );
   // enable depth
   glEnable( GL_DEPTH_TEST );
   // set the front faces of polygons
   glFrontFace( GL_CCW );
   // set fill mode
   glPolygonMode( GL_FRONT_AND_BACK, gFillmode );
   // enable lighting
   glEnable( GL_LIGHTING );
   // enable lighting for front
   glLightModeli( GL_FRONT_AND_BACK, GL_TRUE );
   // material have diffuse and ambient lighting 
   glColorMaterial( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE );
   // enable color
   glEnable( GL_COLOR_MATERIAL );

   // enable light 0
   glEnable( GL_LIGHT0 );

   // setup and enable light 1
   glLightfv( GL_LIGHT1, GL_AMBIENT, gLight1Ambient );
   glLightfv( GL_LIGHT1, GL_DIFFUSE, gLight1Diffuse );
   glLightfv( GL_LIGHT1, GL_SPECULAR, gLight1Specular );
   glEnable( GL_LIGHT1 );

   //glEnable (GL_LINE_SMOOTH);
   glEnable (GL_BLEND);
   glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
   //glHint (GL_LINE_SMOOTH_HINT, GL_DONT_CARE);
}

void reshuffleDisplays()
{
   int N = 0;
   std::deque<GLDisplay*>::iterator i;
   for (i = gDisplays.begin(); i != gDisplays.end(); ++i)
   {
      if ((*i)->Visible())
      {
         ++N;
      }
   }
   
   float height = 1.f / (std::max(N, 1));
   float yVal = - (1 - height);
   float inc = 2 * (height);
   
   for (i = gDisplays.begin(); i != gDisplays.end(); ++i)
   {
      if ((*i)->Visible())
      {
         (*i)->SetY(yVal);
         (*i)->SetH(1);
         if (gStaggered)
            (*i)->SetZ(yVal);
         yVal += inc;
      }
   }
   
   if (gToggle.Toggled())
   {
      gWaterfall.SetY(.95);
      //gWaterfall.SetW(2);
      gWaterfall.SetH(1.8);
      gWaterfall.SetInitialZAngle(-180);
      gWaterfall.SetFade(false);
   }
   else 
   {
      gWaterfall.SetInitialZAngle(-45);
      gWaterfall.SetFade(true);
   }

}

//-----------------------------------------------------------------------------
// Name: reshapeFunc( )
// Desc: called when window size changes
//-----------------------------------------------------------------------------
void reshapeFunc( int w, int h )
{
   // save the new window size
   gWidth = w; gHeight = h;
   
   reshuffleDisplays();
   
   // map the view port to the client area
   glViewport(0,0,w,h);
   //glViewport(0,h/3,w,h/3);
   //glViewport(0,2*h/3,w,h/3);
   // set the matrix mode to project
   glMatrixMode( GL_PROJECTION );
   // load the identity matrix
   glLoadIdentity( );
   // create the viewing frustum
   //gluPerspective( 45.0, (GLfloat) w / (GLfloat) h, 1.0, 300.0 );
   gluPerspective( 90,1, .01, 300.0 );
   // set the matrix mode to modelview
   glMatrixMode( GL_MODELVIEW );
   // load the identity matrix
   glLoadIdentity( );
   // position the view point
   //gluLookAt(0.0f, 3.5f * sin( gEyeY ), 3.5f * cos( gEyeY ), 
   //          0.0f, 0.0f, 0.0f, 0.0f, ( cos( gEyeY ) < 0 ? -1.0f : 1.0f ), 0.0f );
   gluLookAt(0.0f, 0, 1, 
             0.0f, 0.0f, 0.0f, 0.0f, 1, 0.0f );

   //glEnable(GL_SCISSOR_TEST);
   
   //std::cout << "Reshape: w=" << w << "h=" << h << std::endl;
   gAccumulator.SetSamplesPerPixel(std::max((int)(gSamplesPerFrame / w), 1));

   // set the position of the lights
   glLightfv( GL_LIGHT0, GL_POSITION, gLight0Pos );
   glLightfv( GL_LIGHT1, GL_POSITION, gLight1Pos );
   
   //glViewport(0,0,gWidth,gHeight);
   //glScissor(0,0,gWidth,gHeight);
   //glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
}

//-----------------------------------------------------------------------------
// Name: keyboardFunc( )
// Desc: key event
//-----------------------------------------------------------------------------
void keyboardFunc( unsigned char key, int x, int y )
{
   switch( key )
   {
      case 'f':
		{
			if( !gFullscreen )
			{
				gLastWidth = gWidth;
				gLastHeight = gHeight;
				glutFullScreen();
			}
			else
				glutReshapeWindow( gLastWidth, gLastHeight );
			
			gFullscreen = !gFullscreen;
		}
         break;
      case 'h':
         gShowHelp = !gShowHelp;
         break;
      case 'w':
         gFrameRate++;
         gRefreshRate = 1000 / gFrameRate;
         std::cout << "refresh rate: " << gFrameRate << std::endl;
         break;
         
      case 's':
         if (gFrameRate > 1)
         {
            gFrameRate--;
            gRefreshRate = 1000 / gFrameRate;
         }
         std::cout << "refresh rate: " << gFrameRate << std::endl;
         break;
         
      case 'e':
         if (gSamplesPerFrame > 4096)
            gSamplesPerFrame += 512;
         else
            gSamplesPerFrame +=32;
         gAccumulator.SetSamplesPerPixel(std::max((int)(gSamplesPerFrame / gWidth), 1));
         std::cout << "samps/frame: " << gSamplesPerFrame << std::endl;
         break;
         
      case 'd':
         if (gSamplesPerFrame > 4096)
            gSamplesPerFrame -= 512;
         else if (gSamplesPerFrame > 32)
            gSamplesPerFrame -=32;
         gAccumulator.SetSamplesPerPixel(std::max((int)(gSamplesPerFrame / gWidth), 1));
         std::cout << "samps/frame: " << gSamplesPerFrame << std::endl;
         break;
      case '1':
      {
         std::deque<GLDisplay*>::iterator i = std::find(gDisplays.begin(), gDisplays.end(), &gWaveScroll);
         if (i != gDisplays.end())
            gDisplays.erase(i);
         else
            gDisplays.push_back(&gWaveScroll);
         reshuffleDisplays();
      }
         break;
      case '2':
      {
         std::deque<GLDisplay*>::iterator i = std::find(gDisplays.begin(), gDisplays.end(), &gScope);
         if (i != gDisplays.end())
            (*i)->SetVisible(!(*i)->Visible());
         reshuffleDisplays();
      }
         break;
		case '3':
      {
         std::deque<GLDisplay*>::iterator i = std::find(gDisplays.begin(), gDisplays.end(), &gSpectrum);
         if (i != gDisplays.end())
            (*i)->SetVisible(!(*i)->Visible());
         reshuffleDisplays();
      }
			break;
      case '4':
      {
         std::deque<GLDisplay*>::iterator i = std::find(gDisplays.begin(), gDisplays.end(), &gWaterfall);
         if (i != gDisplays.end())
            (*i)->SetVisible(!(*i)->Visible());
         reshuffleDisplays();
      }
			break;
      case 'r':
         gAngleY = 0;
         gAngleX = 0;
         break;
      case 'c':
         cycleColors();
         break;
         
		case 'q':
		{
			exit(0);
		}
			break;
		default:
			break;
   }
	
   // do a reshape since gEyeY might have changed
   reshapeFunc( gWidth, gHeight );
   glutPostRedisplay( );
}

//-----------------------------------------------------------------------------
// Name: mouseFunc( )
// Desc: handles mouse stuff
//-----------------------------------------------------------------------------
void mouseFunc( int button, int state, int x, int y )
{
    if( button == GLUT_LEFT_BUTTON )
    {
        // rotate
       if( state == GLUT_DOWN )
       {
          gLeftButtondown = true;
          gMouseOriginX = x;
          gMouseOriginY = y;   
          
          if (gLoggynessSlider.Hit(x, y))
          {
             gLoggynessSlider.Touch(x, y);
          }
          
          if (gScrollSizeSlider.Hit(x, y))
          {
             gScrollSizeSlider.Touch(x, y);
          }
          
          if (gToggle.Hit(x,y))
          {
             gToggle.Toggle();
          }
       }
       else if (state == GLUT_UP)
       {
          gLoggynessSlider.UnTouch();
          gScrollSizeSlider.UnTouch();
          gLeftButtondown = false;
       }
           // gInc -= gInc_val_mouse;
       else {
            //gInc += gInc_val_mouse;
           
            
       }
    }
    else if ( button == GLUT_RIGHT_BUTTON )
    {
        if( state == GLUT_DOWN )
            gInc += gInc_val_mouse;
        else
            gInc -= gInc_val_mouse;
    }
    else
        gInc = 0.0f;

    glutPostRedisplay( );
}

void motionFunc(int x, int y )
{   
   if (gLeftButtondown)
   {
      if (gLoggynessSlider.Touched())
      {
         gLoggynessSlider.Touch(x,y);
      }
      else if (gScrollSizeSlider.Touched())
      {
         gScrollSizeSlider.Touch(x,y);
      }
      else
      {
         int diffX = x - gMouseOriginX;
         gAngleY +=  diffX / 4.f;
         int diffY = y - gMouseOriginY;
         gAngleX += diffY / 4.f;
         
         std:: cout << gAngleX << " " << gAngleY << "\n";
      }
   }
   
   gMouseOriginX = x;
   gMouseOriginY = y;
}

void displayFunc( )
{
	AudioServer::GetInstance()->GetInput(gBuff, gBufferSize, 0);
   
   fvec vi(gBufferSize);
    memcpy(vi.data_, gBuff, gBufferSize * sizeof(float));
   fvec vo(gBufferSize);
   
   gAutoCorrelation.process(vi, vo);

   float pitch = gAutoCorrelation.pitch();
   pitch = 44100 * (1.f /pitch);
   
   glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
   
   UpdateControls();
   
   glPushMatrix();
   glRotatef( gAngleY, 0, 1.0f, 0.0);
   glRotatef( gAngleX, 1, 0, 0.0);
   
   std::deque<GLDisplay*>::iterator i;
   for (i = gDisplays.begin(); i != gDisplays.end(); ++i)
   {
      if ((*i)->Visible())
      {
         (*i)->Display();
      }
   }
   
   glPopMatrix();
   
   gLoggynessSlider.Draw();
   gScrollSizeSlider.Draw();
   gToggle.Draw();
   
   glColor4f(1,1,1,1);
   std::string s = "Loggyness";
   DrawString(gLoggynessSlider.X() - 20,gLoggynessSlider.Y() - 12, s, GLUT_BITMAP_TIMES_ROMAN_24);
   
   glColor4f(1,1,1,1);
   s = "ScrollSize";
   DrawString(gScrollSizeSlider.X() - 20,gScrollSizeSlider.Y() - 12, s, GLUT_BITMAP_TIMES_ROMAN_24);
  
   glColor4f(1,1,1,1);
   s = "Deluge";
   DrawString(gToggle.X() - 10,gToggle.Y() - 12, s, GLUT_BITMAP_TIMES_ROMAN_24);
   
   stringstream ss;
   ss << "FFT maxbin: " << gFFTPitch;
   s = ss.str();
   DrawString(gWidth - 200, 5, s, GLUT_BITMAP_TIMES_ROMAN_10, 0.125);
   
   ss.str("");
   ss << "Auto Correl.: " << pitch;
   s = ss.str();
   DrawString(gWidth - 200, 30, s, GLUT_BITMAP_TIMES_ROMAN_10, 0.125);

   if (gShowHelp)
   {
      glLineWidth(1);
      s = "Keys: 1, 2, 3, 4 (modes) R (re-orient) F (fullscreen) W/S (refresh rate +/-) Q (quit) H (hide/show me)";
      DrawString(gWidth - 500, gHeight-10, s, GLUT_BITMAP_TIMES_ROMAN_10, 0.07);
   }
   
   glFlush( );
   glutSwapBuffers( );
}

void timerCallback(int v)
{
   //gAngleY += gInc;
   glutPostRedisplay();
   glutTimerFunc(gRefreshRate,timerCallback,0);
}

void createMenu(void){
   //////////
   // MENU //
   //////////
   
   // Create a submenu, this has to be done first.
   int submenid = glutCreateMenu(menu);
   
   // Add sub menu entry
   glutAddMenuEntry("Teapot", 2);
   glutAddMenuEntry("Cube", 3);
   glutAddMenuEntry("Torus", 4);
   
   // Create the menu, this menu becomes the current menu
   int menid = glutCreateMenu(menu);
   
   // Create an entry
   glutAddMenuEntry("Clear", 1);
   
   glutAddSubMenu("Draw", submenid);
   // Create an entry
   glutAddMenuEntry("Quit", 0);
   
   // Let the menu respond on the right mouse button
   glutAttachMenu(GLUT_RIGHT_BUTTON);
   
   
}


void menu(int value){
   if(value == 0){
      exit(0);
   }else{
      gMenuItem = value;
   }
   
   // you would want to redraw now
   glutPostRedisplay();
}

//-----------------------------------------------------------------------------
// Name: idleFunc( )
// Desc: callback from GLUT
//-----------------------------------------------------------------------------
void idleFunc( )
{
   // render the scene
   //glutPostRedisplay( );
}

void cycleColors()
{
   gColorScheme = (gColorScheme + 1) % kNumColorSchemes;
   std::deque<GLDisplay*>::iterator i;
   int count = 0;
   for (i = gDisplays.begin(); i != gDisplays.end(); ++i)
   {
      const unsigned* scheme = qColorSchemes[gColorScheme];
      (*i)->SetColor((scheme[count] & 0xFF0000) / 255.f,
                      (scheme[count] & 0x00FF00) / 255.f,
                      (scheme[count] & 0x0000FF) / 255.f,
                      1);
      count++;
   }
   
}

void DrawString(int x, int y, std::string& s, void* font, float scaleFactor/* = 0.2f*/)
{
  go2d();
   
   glRasterPos3f(10, 10, 0);
   
   glTranslatef(x,y,0); 
   glScalef(scaleFactor,scaleFactor,1);
   
   glColor4f(1,1,1,1);
   for (string::iterator i = s.begin(); i != s.end(); ++i)
   {
      glutStrokeCharacter( GLUT_STROKE_ROMAN, *i );
      //glutBitmapCharacter(font, *i);
   }
   
   go3d();
}


