/* sig-gen.cpp by Adam Somers for Music 256a hw1 */

#include <iostream>

#include "AudioServer.h"
#include "SignalGenerators.h"

#include "chuck_fft.h"

// OpenGL
#if defined(__MACOSX_CORE__)
#include <GLUT/glut.h>
#include <OpenGL/gl.h>
#include <OpenGL/glu.h>
#else
#include <GL/glut.h>
#include <GL/gl.h>
#include <GL/glu.h>
#endif

// gl callbacks
void idleFunc( );
void displayFunc( );
void reshapeFunc( int width, int height );
void keyboardFunc( unsigned char, int, int );
void mouseFunc( int button, int state, int x, int y );
void motionFunc(int x, int y);
void timerCallback(int v);
void initialize_graphics( );
void createMenu(void);
void menu(int value);

// width and height of the window
GLsizei gWidth = 800;
GLsizei gHeight = 600;
GLsizei gLastWidth = gWidth;
GLsizei gLastHeight = gHeight;

GLenum gFillmode = GL_FILL;
GLfloat gLight0Pos[4] = { 2.0f, 1.2f, 4.0f, 1.0f };
GLfloat gLight1Ambient[] = { .2f, .2f, .2f, 1.0f };
GLfloat gLight1Diffuse[] = { 1.0f, 1.0f, 1.0f, 1.0f };
GLfloat gLight1Specular[] = { 1.0f, 1.0f, 1.0f, 1.0f };
GLfloat gLight1Pos[4] = { -2.0f, 0.0f, -4.0f, 1.0f };
// modelview stuff
GLfloat gAngleY = 0.0f;
GLfloat gAngleX = 0.0f;
GLfloat gInc = 0.0f;
GLfloat gEyeY = 0.0f; // 0.2f for stria, otherwise 0

GLboolean gDisplay = TRUE;
GLboolean gFullscreen = FALSE;

const int gBufferSize = 1024;
int gFFTSize = gBufferSize*2;

float gFrameRate = 40; // frames/sec
float gRefreshRate = 1000 / gFrameRate; // ms
float gSamplesPerFrame = 12000;

float gBuff[gBufferSize];
float gWindowBuffer[gBufferSize];
float gFFTBuffer[gBufferSize];
// stores min/max peak pairs
SampleAccumulator::PeakBuffer gPeaks;
// obtains peak data from the signal chain
SampleAccumulator gAccumulator;

bool gScroll = false;
bool gScope = false;
bool gSpectrum = false;

int gMenuItem = 0;
GLfloat gInc_val_mouse = 360 / gFrameRate;

bool gLeftButtondown = false;
int gMouseOriginX = 0;
int gMouseOriginY = 0;

// experimental
int gHopCount = 0;
bool gRings = false;
int gHop = 10;

enum Mode
{
   kScroll = 0,
   kScope,
   kSpectrum,
   
   kNumModes
};

typedef struct _ViewportDescriptor
{
   int x, y, w, h;
   int mode;
} ViewportDescriptor;

std::deque<ViewportDescriptor> gViewports;

int main(int argc, char** argv)
{
	
   memset(gBuff, 0.f, gBufferSize * sizeof(float));
   memset(gFFTBuffer, 0.f, gBufferSize * sizeof(float));

   for (int i = 0; i < gSamplesPerFrame; ++i)
   {
      gPeaks.push_back(std::make_pair(0,0));
   }
	
   // make the transform window
   hanning( gWindowBuffer, gBufferSize );

	std::cout << "Hit q to quit.\n";
	
   RtAudioDriver driver;    // init rtaudio
   
   InputSource i;          // connect input to output through muted accumulator
   gAccumulator.SetInput(&i);
   Multiplier m(&gAccumulator,NULL,0.f);
   AudioServer::GetInstance()->AddClient(&m, 0);

#ifdef __OS_MACOSX__
   //save working dir
   char * cwd = getcwd( NULL, 0 );
#endif
   // initialize GLUT
   glutInit( &argc, argv );
  
#ifdef __OS_MACOSX__
   //restore working dir
   chdir( cwd );
   free( cwd );
#endif
  
   glutInitDisplayMode( GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH );
   glutInitWindowSize( gWidth, gHeight );
   glutInitWindowPosition( 100, 100 );
   glutCreateWindow( "SoundPrism" );
   //glutIdleFunc( idleFunc );
   glutDisplayFunc( displayFunc );
   glutReshapeFunc( reshapeFunc );
   glutKeyboardFunc( keyboardFunc );
   glutMouseFunc( mouseFunc );
   glutMotionFunc(motionFunc);        
   glutTimerFunc(gRefreshRate,timerCallback,0);
   createMenu();
   if( gFullscreen )
      glutFullScreen();
   
   // do our own initialization
   initialize_graphics( );

   glutMainLoop();
	
   for(;;)
      sleep(1);

   return 0;
   }


void initialize_graphics()
{
   // set the GL clear color - use when the color buffer is cleared
   glClearColor( 0.0f, 0.0f,0.0f, 1.0f );
   // set the shading model to 'smooth'
   glShadeModel( GL_SMOOTH );
   // enable depth
   glEnable( GL_DEPTH_TEST );
   // set the front faces of polygons
   glFrontFace( GL_CCW );
   // set fill mode
   glPolygonMode( GL_FRONT_AND_BACK, gFillmode );
   // enable lighting
   glEnable( GL_LIGHTING );
   // enable lighting for front
   glLightModeli( GL_FRONT_AND_BACK, GL_TRUE );
   // material have diffuse and ambient lighting 
   glColorMaterial( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE );
   // enable color
   glEnable( GL_COLOR_MATERIAL );

   // enable light 0
   glEnable( GL_LIGHT0 );

   // setup and enable light 1
   glLightfv( GL_LIGHT1, GL_AMBIENT, gLight1Ambient );
   glLightfv( GL_LIGHT1, GL_DIFFUSE, gLight1Diffuse );
   glLightfv( GL_LIGHT1, GL_SPECULAR, gLight1Specular );
   glEnable( GL_LIGHT1 );

   glEnable (GL_LINE_SMOOTH);
   //glEnable (GL_BLEND);
   //glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
   glHint (GL_LINE_SMOOTH_HINT, GL_DONT_CARE);
}

//-----------------------------------------------------------------------------
// Name: reshapeFunc( )
// Desc: called when window size changes
//-----------------------------------------------------------------------------
void reshapeFunc( int w, int h )
{
   // save the new window size
   gWidth = w; gHeight = h;
   
   
   int yVal = 0;
   
   int height = gHeight / std::max((int)gViewports.size(), 1);
   
   std::deque<ViewportDescriptor>::iterator i;
   for (i = gViewports.begin(); i != gViewports.end(); ++i)
   {
      (*i).w = gWidth;
      (*i).y = yVal;
      (*i).h = height;
      //glViewport((*i).x,(*i).y,(*i).w,(*i).h);
      //glScissor((*i).x,(*i).y,(*i).w,(*i).h);
      yVal += height;
   }
   
   // map the view port to the client area
   //glViewport(0,0,w,h);
   //glViewport(0,h/3,w,h/3);
   //glViewport(0,2*h/3,w,h/3);
   // set the matrix mode to project
   glMatrixMode( GL_PROJECTION );
   // load the identity matrix
   glLoadIdentity( );
   // create the viewing frustum
   gluPerspective( 45.0, (GLfloat) w / (GLfloat) h, 1.0, 300.0 );
   // set the matrix mode to modelview
   glMatrixMode( GL_MODELVIEW );
   // load the identity matrix
   glLoadIdentity( );
   // position the view point
   gluLookAt(0.0f, 3.5f * sin( gEyeY ), 3.5f * cos( gEyeY ), 
             0.0f, 0.0f, 0.0f, 0.0f, ( cos( gEyeY ) < 0 ? -1.0f : 1.0f ), 0.0f );

   glEnable(GL_SCISSOR_TEST);
   
   std::cout << "Reshape: w=" << w << "h=" << h << std::endl;
   gAccumulator.SetSamplesPerPixel(std::max((int)(gSamplesPerFrame / w), 1));

   // set the position of the lights
   glLightfv( GL_LIGHT0, GL_POSITION, gLight0Pos );
   glLightfv( GL_LIGHT1, GL_POSITION, gLight1Pos );
   
   //glViewport(0,0,gWidth,gHeight);
   //glScissor(0,0,gWidth,gHeight);
   //glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
}

void addViewport(int mode)
{
   int nViewports = gViewports.size() + 1;
   int height = gHeight / nViewports;
   ViewportDescriptor desc;
   desc.mode = mode;
   desc.x = 0;
   desc.y = 0;
   desc.w = gWidth;
   desc.h = height;
   std::deque<ViewportDescriptor>::iterator i;
   for (i = gViewports.begin(); i != gViewports.end(); ++i)
   {
      (*i).y += height;
      (*i).h = height;
   }
   gViewports.push_front(desc);   
}

void removeViewport(int mode)
{
   std::deque<ViewportDescriptor>::iterator i;
   for (i = gViewports.begin(); i != gViewports.end(); ++i)
   {
      if ((*i).mode == mode)
      {
         gViewports.erase(i);
         break;
      }
   }
   
   int yVal = 0;
   
   for (i = gViewports.begin(); i != gViewports.end(); ++i)
   {
      (*i).w = gWidth;
      (*i).y = yVal;
      (*i).h = gHeight / gViewports.size();
      yVal += gHeight / gViewports.size();
   }  
}

//-----------------------------------------------------------------------------
// Name: keyboardFunc( )
// Desc: key event
//-----------------------------------------------------------------------------
void keyboardFunc( unsigned char key, int x, int y )
{
   switch( key )
   {
      case 'f':
		{
			if( !gFullscreen )
			{
				gLastWidth = gWidth;
				gLastHeight = gHeight;
				glutFullScreen();
			}
			else
				glutReshapeWindow( gLastWidth, gLastHeight );
			
			gFullscreen = !gFullscreen;
		}
         break;
      case 'w':
         gFrameRate++;
         gRefreshRate = 1000 / gFrameRate;
         std::cout << "refresh rate: " << gFrameRate << std::endl;
         break;
         
      case 's':
         if (gFrameRate > 1)
         {
            gFrameRate--;
            gRefreshRate = 1000 / gFrameRate;
         }
         std::cout << "refresh rate: " << gFrameRate << std::endl;
         break;
         
      case 'e':
         if (gSamplesPerFrame > 4096)
            gSamplesPerFrame += 512;
         else
            gSamplesPerFrame +=32;
         gAccumulator.SetSamplesPerPixel(std::max((int)(gSamplesPerFrame / gWidth), 1));
         std::cout << "samps/frame: " << gSamplesPerFrame << std::endl;
         break;
         
      case 'd':
         if (gSamplesPerFrame > 4096)
            gSamplesPerFrame -= 512;
         else if (gSamplesPerFrame > 32)
            gSamplesPerFrame -=32;
         gAccumulator.SetSamplesPerPixel(std::max((int)(gSamplesPerFrame / gWidth), 1));
         std::cout << "samps/frame: " << gSamplesPerFrame << std::endl;
         break;
      case '1':
         gScroll = !gScroll;
         if (gScroll) // (reshuffle descriptors and add the new descriptor)
         {
            addViewport(kScroll);
         }
         else // remove it and reshuf
         {
            removeViewport(kScroll);
         }
         break;
      case '2':
         gScope = !gScope;
         if (gScope) // (reshuffle descriptors and add the new descriptor)
         {
            addViewport(kScope);
         }
         else // remove it and reshuf
         {
            removeViewport(kScope);
         }
         break;
		case '3':
			gSpectrum = !gSpectrum;
         if (gSpectrum) // (reshuffle descriptors and add the new descriptor)
         {
            addViewport(kSpectrum);
         }
         else // remove it and reshuf
         {
            removeViewport(kSpectrum);
         }
         break;
			break;
         
      case 'r':
         gAngleY = 0;
         gAngleX = 0;
         break;
         
		case 'q':
		{
			exit(0);
		}
			break;
		default:
			break;
   }
	
   // do a reshape since gEyeY might have changed
   reshapeFunc( gWidth, gHeight );
   glutPostRedisplay( );
}

//-----------------------------------------------------------------------------
// Name: mouseFunc( )
// Desc: handles mouse stuff
//-----------------------------------------------------------------------------
void mouseFunc( int button, int state, int x, int y )
{
    if( button == GLUT_LEFT_BUTTON )
    {
        // rotate
       if( state == GLUT_DOWN ){
           gLeftButtondown = true;
          gMouseOriginX = x;
          gMouseOriginY = y;   
    }
           // gInc -= gInc_val_mouse;
       else {
            //gInc += gInc_val_mouse;
           gLeftButtondown = false;
            
       }
    }
    else if ( button == GLUT_RIGHT_BUTTON )
    {
        if( state == GLUT_DOWN )
            gInc += gInc_val_mouse;
        else
            gInc -= gInc_val_mouse;
    }
    else
        gInc = 0.0f;

    glutPostRedisplay( );
}

void motionFunc(int x, int y )
{
   if (gLeftButtondown)
   {
      int diffX = x - gMouseOriginX;
      gAngleY +=  diffX / 4.f;
      gMouseOriginX = x;
      int diffY = y - gMouseOriginY;
      gAngleX += diffY / 4.f;
      gMouseOriginY = y;
   }
}

void drawScroll()
{
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	glPushMatrix( );

	// get the current buffer's peak data
   SampleAccumulator::PeakBuffer peaks = gAccumulator.Get();
   while (!peaks.empty())
   {
      gPeaks.push_back(peaks.front());
      peaks.pop_front();
   }
   
   while (gPeaks.size() > gSamplesPerFrame)
      gPeaks.pop_front();
   
	float x = -2;
	float xInc = fabs((2*x) / (float)gPeaks.size());
   	
	glColor4f(.75, 0, 0, 1);
	
	glLineWidth(1);
	
	glRotatef( gAngleY, 0, 1.0f, 0.0);
   glRotatef( gAngleX, 1, 0, 0.0);
   
	if (!gRings)
	{
		glBegin( GL_LINE_STRIP );
	   SampleAccumulator::PeakBuffer::iterator i;
	   for (i = gPeaks.begin(); i != gPeaks.end(); ++i)
	   {
		  
		  
		  float y0 = (*i).first;
		  float y1 = (*i).second;
		  
		  glVertex3f(x, y0, 0);
		  glVertex3f(x, y1, 0);
		  
		  x+=xInc;
	   }
	   glEnd();
	}
	else
	{
		SampleAccumulator::PeakBuffer::iterator i;
		for (i = gPeaks.begin(); i != gPeaks.end(); ++i)
		{
			float y0 = (*i).first;
			float y1 = (*i).second;
		glBegin( GL_LINE_STRIP );
		for(int i=0;i<35;i++){
			float xcoord = y1 * cos(i*2*PI/10.f);
			float ycoord = y1 * sin(i*2*PI/10.f);	
			glVertex3f(x, ycoord, xcoord);
		}
			x+=xInc;
		glEnd();
		}
	}
	
	
   glPopMatrix( );   
}

bool gTriggered = false;

void drawScope()
{   
   int delay = 0;
   
   for (int i = 0; i < gBufferSize - 1; ++i)
   {
      if (gBuff[i] < 0.0001 && gBuff[i+1] >= 0.0001)
      {
         gTriggered = true;
         break;
      }
      delay++;
   }
   
   if (!gTriggered)
      return;

	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	glPushMatrix( );
	
	float x = -2;
	float xInc =fabs((2*x) / (float)gBufferSize);

	glColor4f(.75, 0, 0, 1);
	
	glLineWidth(1);
	
	glRotatef( gAngleY, 0, 1.0f, 0.0);
   glRotatef( gAngleX, 1, 0, 0.0);

   if (!gRings)
   {
      glBegin( GL_LINE_STRIP );
      for (int i = 0; i < gBufferSize; ++i)
      {
         if (i < delay)
            continue;

         float y = gBuff[i];
         glVertex3f(x, y, 0);
         x+=xInc;
      }
      glEnd();
   }
   else
   {
   
      for (int i = 0; i < gBufferSize; ++i)
      {
         if (i < delay)
            continue;
            
   		float y = gBuff[i];
   		
         if (!gRings)
         {
            glVertex3f(x, y, 0);
         }
         else
         {
            glBegin( GL_LINE_LOOP );	
            for(int i=0;i<35;i++){
               float xcoord = y * cos(i*2*PI/10.f);
               float ycoord = y * sin(i*2*PI/10.f);	
               glVertex3f(x, ycoord, xcoord);
            }        
            glEnd();
         }
   		x+=xInc;
   	}
   }
   
	glPopMatrix( );

   gTriggered = false;
   
}

void drawFFT()
{
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	glPushMatrix( );
	
	float x = -2;
	float xInc =fabs((2*x) / (float)gBufferSize) * 2;
	
	glColor4f(.75, 0, 0, 1);
	
	glLineWidth(1);
	
	glRotatef( gAngleY, 0, 1.0f, 0.0);
	glRotatef( gAngleX, 1, 0, 0.0);
	
   memcpy( gFFTBuffer, gBuff, gBufferSize * sizeof(float));
   
	apply_window( (float*)gFFTBuffer, gWindowBuffer, gBufferSize );
	rfft( (float *)gFFTBuffer, gBufferSize/2, FFT_FORWARD );
	// cast to complex
	complex * cbuf = (complex *)gFFTBuffer;
	
	glBegin( GL_LINE_STRIP );
	for (int i = 0; i < gBufferSize/2; ++i)
	{
		float y = ::pow( 25 * cmp_abs( cbuf[i] ), .5 );
		glVertex3f(x, y, 0);
		x+=xInc;
	}
	glEnd();
	
	glPopMatrix( );	
}

void drawSpheres()
{
   glPushMatrix();
   glRotatef( gAngleY / 2.f, 0.0f, 1.0f, 0.0);
   glLineWidth(1);
   glutSolidSphere(.5, 20, 20);
   
   glTranslatef( .75, 0, 0);
   glPushMatrix();
   glRotatef(45, 1, 0, 0.0);
   glRotatef( gAngleY, 0, 0, 1);
   glColor4f(1.0, 0.0, 1.0, 1.0);
   glutSolidSphere(.25, 20, 20);
   glPopMatrix();
   
   glRotatef(gAngleY, 1, 0,0);
   glTranslatef( 0, .375, 0);
   glColor4f(1.0, 1.0, 0.0, 1.0);
   glutSolidSphere(.125, 20, 20);
   
   glPopMatrix();
}

void displayFunc( )
{
	AudioServer::GetInstance()->GetInput(gBuff, gBufferSize, 0);
   
   if (gViewports.size() == 0)
   {
      glViewport(0,0,gWidth,gHeight);
      glScissor(0,0,gWidth,gHeight);
      glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
      drawSpheres();
   }

   std::deque<ViewportDescriptor>::iterator i;
   for (i = gViewports.begin(); i != gViewports.end(); ++i)
   {
      glViewport((*i).x,(*i).y,(*i).w,(*i).h);
      glScissor ((*i).x,(*i).y,(*i).w,(*i).h);
      int mode = (*i).mode;
      switch (mode)
      {
         case kScroll:
            drawScroll();
            break;
         case kScope:
            drawScope();
            break;
         case kSpectrum:
            drawFFT();
            break;
      }
   }
   
   glFlush( );
   glutSwapBuffers( );
}

void timerCallback(int v)
{
   //gAngleY += gInc;
   glutPostRedisplay();
   glutTimerFunc(gRefreshRate,timerCallback,0);
}

void createMenu(void){
   //////////
   // MENU //
   //////////
   
   // Create a submenu, this has to be done first.
   int submenid = glutCreateMenu(menu);
   
   // Add sub menu entry
   glutAddMenuEntry("Teapot", 2);
   glutAddMenuEntry("Cube", 3);
   glutAddMenuEntry("Torus", 4);
   
   // Create the menu, this menu becomes the current menu
   int menid = glutCreateMenu(menu);
   
   // Create an entry
   glutAddMenuEntry("Clear", 1);
   
   glutAddSubMenu("Draw", submenid);
   // Create an entry
   glutAddMenuEntry("Quit", 0);
   
   // Let the menu respond on the right mouse button
   glutAttachMenu(GLUT_RIGHT_BUTTON);
   
   
}


void menu(int value){
   if(value == 0){
      exit(0);
   }else{
      gMenuItem = value;
   }
   
   // you would want to redraw now
   glutPostRedisplay();
}

//-----------------------------------------------------------------------------
// Name: idleFunc( )
// Desc: callback from GLUT
//-----------------------------------------------------------------------------
void idleFunc( )
{
   // render the scene
   //glutPostRedisplay( );
}
